/*  OpenTally: Open-source election vote counting
 *  Copyright © 2021 Lee Yingtong Li (RunasSudo)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use super::{Assign, From, Number};

use derive_more::Display;
use num_traits::{Num, One, Zero};

use std::cmp::{Ord, Ordering, PartialEq, PartialOrd};
use std::num::ParseIntError;
use std::ops;

type ImplType = f64;

#[derive(Clone, Display, PartialEq, PartialOrd)]
pub struct NativeFloat64(ImplType);

impl Number for NativeFloat64 {
	fn new() -> Self { Self(0.0) }
	
	fn describe() -> String { "--numbers float64".to_string() }
	
	fn pow_assign(&mut self, exponent: i32) {
		self.0 = self.0.powi(exponent);
	}
	
	fn floor_mut(&mut self, dps: usize) {
		let factor = 10.0_f64.powi(dps as i32);
		self.0 = (self.0 * factor).floor() / factor;
	}
	
	fn ceil_mut(&mut self, dps: usize) {
		let factor = 10.0_f64.powi(dps as i32);
		self.0 = (self.0 * factor).ceil() / factor;
	}
}

impl Num for NativeFloat64 {
	type FromStrRadixErr = ParseIntError;
	fn from_str_radix(str: &str, radix: u32) -> Result<Self, Self::FromStrRadixErr> {
		match i64::from_str_radix(str, radix) {
			Ok(value) => Ok(Self(value as ImplType)),
			Err(err) => Err(err)
		}
	}
}

impl Assign for NativeFloat64 {
	fn assign(&mut self, src: Self) { self.0 = src.0; }
}

impl Assign<&NativeFloat64> for NativeFloat64 {
	fn assign(&mut self, src: &NativeFloat64) { self.0 = src.0; }
}

impl From<usize> for NativeFloat64 {
	fn from(n: usize) -> Self { Self(n as ImplType) }
}

impl From<f64> for NativeFloat64 {
	fn from(n: f64) -> Self { Self(n as ImplType) }
}

impl One for NativeFloat64 {
	fn one() -> Self { Self(1.0) }
}

impl Zero for NativeFloat64 {
	fn zero() -> Self { Self::new() }
	fn is_zero(&self) -> bool { self.0.is_zero() }
}

impl Eq for NativeFloat64 {}
impl Ord for NativeFloat64 {
	fn cmp(&self, other: &Self) -> Ordering { self.0.partial_cmp(&other.0).unwrap() }
}

impl ops::Neg for NativeFloat64 {
	type Output = NativeFloat64;
	fn neg(self) -> Self::Output { Self(-self.0) }
}

impl ops::Add for NativeFloat64 {
	type Output = NativeFloat64;
	fn add(self, _rhs: Self) -> Self::Output {
		todo!()
	}
}

impl ops::Sub for NativeFloat64 {
	type Output = NativeFloat64;
	fn sub(self, _rhs: Self) -> Self::Output {
		todo!()
	}
}

impl ops::Mul for NativeFloat64 {
	type Output = NativeFloat64;
	fn mul(self, _rhs: Self) -> Self::Output {
		todo!()
	}
}

impl ops::Div for NativeFloat64 {
	type Output = NativeFloat64;
	fn div(self, _rhs: Self) -> Self::Output {
		todo!()
	}
}

impl ops::Rem for NativeFloat64 {
	type Output = NativeFloat64;
	fn rem(self, _rhs: Self) -> Self::Output {
		todo!()
	}
}

impl ops::Add<&NativeFloat64> for NativeFloat64 {
	type Output = NativeFloat64;
	fn add(self, rhs: &NativeFloat64) -> Self::Output { Self(self.0 + &rhs.0) }
}

impl ops::Sub<&NativeFloat64> for NativeFloat64 {
	type Output = NativeFloat64;
	fn sub(self, rhs: &NativeFloat64) -> Self::Output { Self(self.0 - &rhs.0) }
}

impl ops::Mul<&NativeFloat64> for NativeFloat64 {
	type Output = NativeFloat64;
	fn mul(self, rhs: &NativeFloat64) -> Self::Output { NativeFloat64(self.0 * &rhs.0) }
}

impl ops::Div<&NativeFloat64> for NativeFloat64 {
	type Output = NativeFloat64;
	fn div(self, rhs: &NativeFloat64) -> Self::Output { NativeFloat64(self.0 / &rhs.0) }
}

impl ops::Rem<&NativeFloat64> for NativeFloat64 {
	type Output = NativeFloat64;
	fn rem(self, _rhs: &NativeFloat64) -> Self::Output {
		todo!()
	}
}

impl ops::AddAssign for NativeFloat64 {
	fn add_assign(&mut self, rhs: Self) { self.0 += rhs.0; }
}

impl ops::SubAssign for NativeFloat64 {
	fn sub_assign(&mut self, rhs: Self) { self.0 -= rhs.0; }
}

impl ops::MulAssign for NativeFloat64 {
	fn mul_assign(&mut self, rhs: Self) { self.0 *= rhs.0; }
}

impl ops::DivAssign for NativeFloat64 {
	fn div_assign(&mut self, rhs: Self) { self.0 /= &rhs.0; }
}

impl ops::RemAssign for NativeFloat64 {
	fn rem_assign(&mut self, _rhs: Self) {
		todo!()
	}
}

impl ops::AddAssign<&NativeFloat64> for NativeFloat64 {
	fn add_assign(&mut self, rhs: &NativeFloat64) { self.0 += &rhs.0; }
}

impl ops::SubAssign<&NativeFloat64> for NativeFloat64 {
	fn sub_assign(&mut self, rhs: &NativeFloat64) { self.0 -= &rhs.0; }
}

impl ops::MulAssign<&NativeFloat64> for NativeFloat64 {
	fn mul_assign(&mut self, rhs: &NativeFloat64) { self.0 *= &rhs.0; }
}

impl ops::DivAssign<&NativeFloat64> for NativeFloat64 {
	fn div_assign(&mut self, _rhs: &NativeFloat64) {
		todo!()
	}
}

impl ops::RemAssign<&NativeFloat64> for NativeFloat64 {
	fn rem_assign(&mut self, _rhs: &NativeFloat64) {
		todo!()
	}
}

impl ops::Neg for &NativeFloat64 {
	type Output = NativeFloat64;
	fn neg(self) -> Self::Output { NativeFloat64(-&self.0) }
}

impl ops::Add<&NativeFloat64> for &NativeFloat64 {
	type Output = NativeFloat64;
	fn add(self, _rhs: &NativeFloat64) -> Self::Output {
		todo!()
	}
}

impl ops::Sub<&NativeFloat64> for &NativeFloat64 {
	type Output = NativeFloat64;
	fn sub(self, rhs: &NativeFloat64) -> Self::Output { NativeFloat64(&self.0 - &rhs.0) }
}

impl ops::Mul<&NativeFloat64> for &NativeFloat64 {
	type Output = NativeFloat64;
	fn mul(self, _rhs: &NativeFloat64) -> Self::Output {
		todo!()
	}
}

impl ops::Div<&NativeFloat64> for &NativeFloat64 {
	type Output = NativeFloat64;
	fn div(self, rhs: &NativeFloat64) -> Self::Output { NativeFloat64(&self.0 / &rhs.0) }
}

impl ops::Rem<&NativeFloat64> for &NativeFloat64 {
	type Output = NativeFloat64;
	fn rem(self, _rhs: &NativeFloat64) -> Self::Output {
		todo!()
	}
}

/*
impl ops::Add<&&NativeFloat> for &NativeFloat {
	
}

impl ops::Sub<&&NativeFloat> for &NativeFloat {
	
}

impl ops::Mul<&&NativeFloat> for &NativeFloat {
	
}

impl ops::Div<&&NativeFloat> for &NativeFloat {
	
}

impl ops::Rem<&&NativeFloat> for &NativeFloat {
	
}
*/
